@@ -1,3 +1,6 @@
+import logging
+
+
 class MagicComponent:
     """
         To automagically retrieve variables defined in your base robot
@@ -6,8 +9,8 @@ class MagicComponent:
             class MyComponent:
             
                 # other variables 'imported' automatically from MagicRobot
-                elevator_motor = Talon
-                other_component = MyOtherComponent
+                elevator_motor: Talon
+                other_component: MyOtherComponent
                 
                 ...
                 
@@ -27,6 +30,8 @@ class MagicComponent:
                   provided for documentation's sake
     """
 
+    logger: logging.Logger
+
     def setup(self):
         """
             This function is called after ``createObjects`` has been called in
