@@ -1,19 +1,23 @@
 import contextlib
 import inspect
 import logging
+import typing
 
+from typing import Any, Callable, Dict, List, Tuple
+
+import hal
 import wpilib
 
-from robotpy_ext.misc import NotifierDelay
+from networktables import NetworkTables, NetworkTableEntry
+
+# from wpilib.shuffleboard import Shuffleboard
+
 from robotpy_ext.autonomous import AutonomousModeSelector
+from robotpy_ext.misc import NotifierDelay
+from robotpy_ext.misc.simple_watchdog import SimpleWatchdog
 
-from robotpy_ext.misc.orderedclass import OrderedClass
-from robotpy_ext.misc.annotations import get_class_annotations
-
-from networktables import NetworkTables
-
-from .magic_tunable import setup_tunables, _TunableProperty, collect_feedbacks
-from .magic_reset import will_reset_to
+from .magic_tunable import setup_tunables, tunable, collect_feedbacks
+from .magic_reset import collect_resets
 
 __all__ = ["MagicRobot"]
 
@@ -22,7 +26,7 @@ class MagicInjectError(ValueError):
     pass
 
 
-class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
+class MagicRobot(wpilib.RobotBase):
     """
         Robots that use the MagicBot framework should use this as their
         base robot class. If you use this as your base, you must
@@ -33,8 +37,8 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
         MagicRobot uses the :class:`.AutonomousModeSelector` to allow you
         to define multiple autonomous modes and to select one of them via
-        the SmartDashboard/SFX.
-        
+        the SmartDashboard/Shuffleboard.
+
         MagicRobot will set the following NetworkTables variables
         automatically:
         
@@ -51,19 +55,37 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
     #: uncaught exceptions be reported?
     error_report_interval = 0.5
 
+    #: A Python logging object that you can use to send messages to the log.
+    #: It is recommended to use this instead of print statements.
+    logger = logging.getLogger("robot")
+
+    #: If True, teleopPeriodic will be called in autonomous mode
+    use_teleop_in_autonomous = False
+
+    def __init__(self) -> None:
+        super().__init__()
+
+        self._exclude_from_injection = ["logger"]
+
+        self.__last_error_report = -10
+
+        self._components: List[Tuple[str, Any]] = []
+        self._feedbacks: List[Tuple[Callable[[], Any], NetworkTableEntry]] = []
+        self._reset_components: List[Tuple[Dict[str, Any], Any]] = []
+
+        self.__done = False
+
+    def _simulationInit(self):
+        pass
+
+    def _simulationPeriodic(self):
+        pass
+
     def robotInit(self):
         """
             .. warning:: Internal API, don't override; use :meth:`createObjects` instead
         """
 
-        self._exclude_from_injection = ["logger", "members"]
-
-        self.__last_error_report = -10
-
-        self._components = []
-        self._feedbacks = []
-        self._reset_components = []
-
         # Create the user's objects and stuff here
         self.createObjects()
 
@@ -74,8 +96,25 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
         self._create_components()
 
         self.__nt = NetworkTables.getTable("/robot")
+
+        self.__nt_put_is_ds_attached = self.__nt.getEntry("is_ds_attached").setBoolean
+        self.__nt_put_mode = self.__nt.getEntry("mode").setString
+
         self.__nt.putBoolean("is_simulation", self.isSimulation())
-        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
+        self.__nt_put_is_ds_attached(self.ds.isDSAttached())
+
+        # cache these
+        self.__sd_update = wpilib.SmartDashboard.updateValues
+        self.__lv_update = wpilib.LiveWindow.getInstance().updateValues
+        # self.__sf_update = Shuffleboard.update
+
+        self.watchdog = SimpleWatchdog(self.control_loop_wait_time)
+
+        self.__periodics = [(self.robotPeriodic, "robotPeriodic()")]
+
+        if self.isSimulation():
+            self._simulationInit()
+            self.__periodics.append((self._simulationPeriodic, "simulationPeriodic()"))
 
     def createObjects(self):
         """
@@ -98,6 +137,20 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
         """
         raise NotImplementedError
 
+    def autonomousInit(self):
+        """Initialization code for autonomous mode may go here.
+
+        Users may override this method for initialization code which
+        will be called each time the robot enters autonomous mode,
+        regardless of the selected autonomous mode.
+
+        This can be useful for code that must be run at the beginning of a match.
+
+        .. note:: The ``on_enable`` functions of all components are
+                  called before this function is called.
+        """
+        pass
+
     def teleopInit(self):
         """
             Initialization code for teleop control code may go here.
@@ -119,11 +172,15 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
             This code executes before the ``execute`` functions of all
             components are called.
+
+            .. note:: If you want this function to be called in autonomous
+                      mode, set ``use_teleop_in_autonomous`` to True in your
+                      robot class.
         """
         func = self.teleopPeriodic.__func__
         if not hasattr(func, "firstRun"):
             self.logger.warning(
-                "Default MagicRobot.teleopPeriodic() method... Overload me!"
+                "Default MagicRobot.teleopPeriodic() method... Override me!"
             )
             func.firstRun = False
 
@@ -152,7 +209,7 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
         func = self.disabledPeriodic.__func__
         if not hasattr(func, "firstRun"):
             self.logger.warning(
-                "Default MagicRobot.disabledPeriodic() method... Overload me!"
+                "Default MagicRobot.disabledPeriodic() method... Override me!"
             )
             func.firstRun = False
 
@@ -179,12 +236,18 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
             You may use it for any code you need to run
             during all modes of the robot (e.g NetworkTables updates)
 
-            The default implementation will update SmartDashboard and LiveWindow.
+            The default implementation will update
+            SmartDashboard, LiveWindow and Shuffleboard.
         """
-        wpilib.SmartDashboard.updateValues()
-        wpilib.LiveWindow.updateValues()
+        watchdog = self.watchdog
+        self.__sd_update()
+        watchdog.addEpoch("SmartDashboard")
+        self.__lv_update()
+        watchdog.addEpoch("LiveWindow")
+        # self.__sf_update()
+        # watchdog.addEpoch("Shuffleboard")
 
-    def onException(self, forceReport=False):
+    def onException(self, forceReport: bool = False) -> None:
         """
             This function must *only* be called when an unexpected exception
             has occurred that would otherwise crash the robot code. Use this
@@ -236,7 +299,7 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
         self.__last_error_report = now
 
     @contextlib.contextmanager
-    def consumeExceptions(self, forceReport=False):
+    def consumeExceptions(self, forceReport: bool = False):
         """
             This returns a context manager which will consume any uncaught
             exceptions that might otherwise crash the robot.
@@ -268,6 +331,34 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
     # Internal API
     #
 
+    def startCompetition(self) -> None:
+        """
+        This runs the mode-switching loop.
+
+        .. warning:: Internal API, don't override
+        """
+
+        # TODO: usage reporting?
+        self.robotInit()
+
+        # Tell the DS the robot is ready to be enabled
+        hal.observeUserProgramStarting()
+
+        while not self.__done:
+            isEnabled, isAutonomous, isTest = self.getControlState()
+
+            if not isEnabled:
+                self._disabled()
+            elif isAutonomous:
+                self.autonomous()
+            elif isTest:
+                self._test()
+            else:
+                self._operatorControl()
+
+    def endCompetition(self) -> None:
+        self.__done = True
+
     def autonomous(self):
         """
             MagicRobot will do The Right Thing and automatically load all
@@ -276,21 +367,33 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
             .. warning:: Internal API, don't override
         """
 
-        self.__nt.putString("mode", "auto")
-        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
-        wpilib.LiveWindow.setEnabled(False)
+        self.__nt_put_mode("auto")
+        self.__nt_put_is_ds_attached(self.ds.isDSAttached())
 
         self._on_mode_enable_components()
 
+        try:
+            self.autonomousInit()
+        except:
+            self.onException(forceReport=True)
+
+        auto_functions = (self._execute_components, self._update_feedback,) + tuple(
+            p[0] for p in self.__periodics
+        )
+
+        if self.use_teleop_in_autonomous:
+            auto_functions = (self.teleopPeriodic,) + auto_functions
+
         self._automodes.run(
             self.control_loop_wait_time,
-            (self._execute_components, self._update_feedback, self.robotPeriodic),
+            auto_functions,
             self.onException,
+            watchdog=self.watchdog,
         )
 
         self._on_mode_disable_components()
 
-    def disabled(self):
+    def _disabled(self):
         """
             This function is called in disabled mode. You should not
             override this function; rather, you should override the
@@ -298,33 +401,44 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
             .. warning:: Internal API, don't override
         """
+        watchdog = self.watchdog
+        watchdog.reset()
 
-        self.__nt.putString("mode", "disabled")
+        self.__nt_put_mode("disabled")
         ds_attached = None
-        wpilib.LiveWindow.setEnabled(False)
 
         self._on_mode_disable_components()
         try:
             self.disabledInit()
         except:
             self.onException(forceReport=True)
+        watchdog.addEpoch("disabledInit()")
 
         with NotifierDelay(self.control_loop_wait_time) as delay:
-            while self.isDisabled():
+            while not self.__done and self.isDisabled():
                 if ds_attached != self.ds.isDSAttached():
                     ds_attached = not ds_attached
-                    self.__nt.putBoolean("is_ds_attached", ds_attached)
+                    self.__nt_put_is_ds_attached(ds_attached)
 
+                hal.observeUserProgramDisabled()
                 try:
                     self.disabledPeriodic()
                 except:
                     self.onException()
+                watchdog.addEpoch("disabledPeriodic()")
 
                 self._update_feedback()
-                self.robotPeriodic()
-                delay.wait()
+                for periodic, name in self.__periodics:
+                    periodic()
+                    watchdog.addEpoch(name)
+                # watchdog.disable()
 
-    def operatorControl(self):
+                watchdog.printIfExpired()
+
+                delay.wait()
+                watchdog.reset()
+
+    def _operatorControl(self):
         """
             This function is called in teleoperated mode. You should not
             override this function; rather, you should override the
@@ -332,12 +446,13 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
             .. warning:: Internal API, don't override
         """
+        watchdog = self.watchdog
+        watchdog.reset()
 
-        self.__nt.putString("mode", "teleop")
+        self.__nt_put_mode("teleop")
         # don't need to update this during teleop -- presumably will switch
         # modes when ds is no longer attached
-        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
-        wpilib.LiveWindow.setEnabled(False)
+        self.__nt_put_is_ds_attached(self.ds.isDSAttached())
 
         # initialize things
         self._on_mode_enable_components()
@@ -346,60 +461,92 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
             self.teleopInit()
         except:
             self.onException(forceReport=True)
+        watchdog.addEpoch("teleopInit()")
+
+        observe = hal.observeUserProgramTeleop
 
         with NotifierDelay(self.control_loop_wait_time) as delay:
-            while self.isOperatorControl() and self.isEnabled():
+            while not self.__done and self.isOperatorControlEnabled():
+                observe()
                 try:
                     self.teleopPeriodic()
                 except:
                     self.onException()
+                watchdog.addEpoch("teleopPeriodic()")
 
                 self._execute_components()
+
                 self._update_feedback()
-                self.robotPeriodic()
+                for periodic, name in self.__periodics:
+                    periodic()
+                    watchdog.addEpoch(name)
+                # watchdog.disable()
+
+                watchdog.printIfExpired()
 
                 delay.wait()
+                watchdog.reset()
 
         self._on_mode_disable_components()
 
-    def test(self):
+    def _test(self):
         """Called when the robot is in test mode"""
+        watchdog = self.watchdog
+        watchdog.reset()
 
-        self.__nt.putString("mode", "test")
-        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
-        wpilib.LiveWindow.setEnabled(True)
+        self.__nt_put_mode("test")
+        self.__nt_put_is_ds_attached(self.ds.isDSAttached())
+
+        lw = wpilib.LiveWindow.getInstance()
+        lw.setEnabled(True)
+        # Shuffleboard.enableActuatorWidgets()
 
         try:
             self.testInit()
         except:
             self.onException(forceReport=True)
+        watchdog.addEpoch("testInit()")
 
         with NotifierDelay(self.control_loop_wait_time) as delay:
-            while self.isTest() and self.isEnabled():
+            while not self.__done and self.isTest() and self.isEnabled():
+                hal.observeUserProgramTest()
                 try:
                     self.testPeriodic()
                 except:
                     self.onException()
+                watchdog.addEpoch("testPeriodic()")
 
                 self._update_feedback()
-                self.robotPeriodic()
+                for periodic, name in self.__periodics:
+                    periodic()
+                    watchdog.addEpoch(name)
+                # watchdog.disable()
+
+                watchdog.printIfExpired()
+
                 delay.wait()
+                watchdog.reset()
+
+        lw.setEnabled(False)
+        # Shuffleboard.disableActuatorWidgets()
 
     def _on_mode_enable_components(self):
         # initialize things
-        for component in self._components:
-            if hasattr(component, "on_enable"):
+        for _, component in self._components:
+            on_enable = getattr(component, "on_enable", None)
+            if on_enable is not None:
                 try:
-                    component.on_enable()
+                    on_enable()
                 except:
                     self.onException(forceReport=True)
 
     def _on_mode_disable_components(self):
         # deinitialize things
-        for component in self._components:
-            if hasattr(component, "on_disable"):
+        for _, component in self._components:
+            on_disable = getattr(component, "on_disable", None)
+            if on_disable is not None:
                 try:
-                    component.on_disable()
+                    on_disable()
                 except:
                     self.onException(forceReport=True)
 
@@ -418,26 +565,23 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
         self.logger.info("Creating magic components")
 
         # Identify all of the types, and create them
-        cls = self.__class__
+        cls = type(self)
 
         # - Iterate over class variables with type annotations
-        for m, ctyp in get_class_annotations(cls).items():
+        # .. this hack is necessary for pybind11 based modules
+        class FakeModule:
+            pass
+
+        import sys
+
+        sys.modules["pybind11_builtins"] = FakeModule()
+
+        for m, ctyp in typing.get_type_hints(cls).items():
             # Ignore private variables
             if m.startswith("_"):
                 continue
 
-            if hasattr(cls, m):
-                attr = getattr(cls, m)
-                # If the value given to the variable is an instance of a type and isn't a property
-                # raise an error. No double declaring types, e.g foo: type = type
-                if isinstance(attr, type) and not isinstance(attr, property):
-                    raise ValueError(
-                        "%s.%s has two type declarations" % (cls.__name__, m)
-                    )
-                # Otherwise, skip this set class variable
-                continue
-
-            # If the variable has been assigned in __init__ or createObjects, skip it
+            # If the variable has been set, skip it
             if hasattr(self, m):
                 continue
 
@@ -453,20 +597,6 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
             # Store for later
             components.append((m, component))
 
-        # - Iterate over set class variables
-        for m in self.members:
-            if m.startswith("_") or isinstance(getattr(cls, m, None), _TunableProperty):
-                continue
-
-            ctyp = getattr(self, m)
-            if not isinstance(ctyp, type):
-                continue
-
-            component = self._create_component(m, ctyp)
-
-            # Store for later
-            components.append((m, component))
-
         # Collect attributes of this robot that are injectable
         self._injectables = {}
 
@@ -474,7 +604,7 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
             if (
                 n.startswith("_")
                 or n in self._exclude_from_injection
-                or isinstance(getattr(cls, n, None), _TunableProperty)
+                or isinstance(getattr(cls, n, None), tunable)
             ):
                 continue
 
@@ -489,9 +619,7 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
         # For each new component, perform magic injection
         for cname, component in components:
-            self._components.append(component)
             setup_tunables(component, cname, "components")
-            self._feedbacks += collect_feedbacks(component, cname, "components")
             self._setup_vars(cname, component)
             self._setup_reset_vars(component)
 
@@ -507,15 +635,20 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
         # Call setup functions for components
         for cname, component in components:
-            if hasattr(component, "setup"):
-                component.setup()
+            setup = getattr(component, "setup", None)
+            if setup is not None:
+                setup()
+            # ... and grab all the feedback methods
+            self._feedbacks += collect_feedbacks(component, cname, "components")
 
         # Call setup functions for autonomous modes
         for mode in self._automodes.modes.values():
             if hasattr(mode, "setup"):
                 mode.setup()
 
-    def _create_component(self, name, ctyp):
+        self._components = components
+
+    def _create_component(self, name: str, ctyp: type):
         # Create instance, set it on self
         component = ctyp()
         setattr(self, name, component)
@@ -534,34 +667,28 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
         return component
 
-    def _setup_vars(self, cname, component):
+    def _setup_vars(self, cname: str, component):
 
         self.logger.debug("Injecting magic variables into %s", cname)
 
         component_type = type(component)
 
         # Iterate over variables with type annotations
-        for n, inject_type in get_class_annotations(component_type).items():
+        for n, inject_type in typing.get_type_hints(component_type).items():
 
             # If the variable is private ignore it
             if n.startswith("_"):
                 continue
 
-            if hasattr(component_type, n):
-                attr = getattr(component_type, n)
-                # If the value given to the variable is an instance of a type and isn't a property
-                # raise an error. No double declaring types, e.g foo: type = type
-                if isinstance(attr, type) and not isinstance(attr, property):
-                    raise MagicInjectError(
-                        "%s.%s has two type declarations" % (component_type.__name__, n)
-                    )
-                # Otherwise, skip this set class variable
-                continue
-
-            # If the variable has been assigned in __init__, skip it
+            # If the variable has been set, skip it
             if hasattr(component, n):
                 continue
 
+            # Check for generic types from the typing module
+            origin = getattr(inject_type, "__origin__", None)
+            if origin is not None:
+                inject_type = origin
+
             # If the type is not actually a type, give a meaningful error
             if not isinstance(inject_type, type):
                 raise TypeError(
@@ -571,23 +698,7 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
 
             self._inject(n, inject_type, cname, component)
 
-        # Iterate over static variables
-        for n in dir(component):
-            # If the variable is private or a proprty, don't inject
-            if n.startswith("_") or isinstance(
-                getattr(component_type, n, True), property
-            ):
-                continue
-
-            inject_type = getattr(component, n)
-
-            # If the value assigned isn't a type, don't inject
-            if not isinstance(inject_type, type):
-                continue
-
-            self._inject(n, inject_type, cname, component)
-
-    def _inject(self, n, inject_type, cname, component):
+    def _inject(self, n: str, inject_type: type, cname: str, component):
         # Retrieve injectable object
         injectable = self._injectables.get(n)
         if injectable is None:
@@ -613,15 +724,7 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
         self.logger.debug("-> %s as %s.%s", injectable, cname, n)
 
     def _setup_reset_vars(self, component):
-        reset_dict = {}
-
-        for n in dir(component):
-            if isinstance(getattr(type(component), n, True), property):
-                continue
-
-            a = getattr(component, n, None)
-            if isinstance(a, will_reset_to):
-                reset_dict[n] = a.default
+        reset_dict = collect_resets(type(component))
 
         if reset_dict:
             component.__dict__.update(reset_dict)
@@ -635,13 +738,15 @@ class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
                 self.onException()
                 continue
             entry.setValue(value)
+        self.watchdog.addEpoch("@magicbot.feedback")
 
     def _execute_components(self):
-        for component in self._components:
+        for name, component in self._components:
             try:
                 component.execute()
             except:
                 self.onException()
+            self.watchdog.addEpoch(name)
 
         for reset_dict, component in self._reset_components:
             component.__dict__.update(reset_dict)
