@@ -1,15 +1,17 @@
-from glob import glob
 import importlib
 import inspect
-import os
-
 import logging
+import os
+from glob import glob
+from typing import Union
 
-logger = logging.getLogger("autonomous")
+import hal
+import wpilib
 
 from ..misc.precise_delay import NotifierDelay
+from ..misc.simple_watchdog import SimpleWatchdog
 
-import wpilib
+logger = logging.getLogger("autonomous")
 
 
 class AutonomousModeSelector:
@@ -17,8 +19,8 @@ class AutonomousModeSelector:
         This object loads all modules in a specified python package, and tries
         to automatically discover autonomous modes from them. Each module is 
         added to a ``SendableChooser`` object, which allows the user to select
-        one of them via the SmartDashboard/SFX.
-        
+        one of them via SmartDashboard.
+
         Autonomous mode objects must implement the following functions:
         
         - ``on_enable`` - Called when autonomous mode is initially enabled
@@ -30,30 +32,25 @@ class AutonomousModeSelector:
         - ``MODE_NAME`` - The name of the autonomous mode to display to users
         - ``DISABLED`` - If True, don't allow this mode to be selected
         - ``DEFAULT`` - If True, this is the default autonomous mode selected
-        
-        AutonomousModeSelector can be easily used from either ``IterativeRobot``
-        or from ``SampleRobot``. For ``IterativeRobot``, the simplest usage would
-        look like so::
-        
-            class MyRobot(wpilib.IterativeRobot):
-            
+
+        Here is an example of using ``AutonomousModeSelector`` in ``TimedRobot``:
+
+        .. code-block:: python
+
+            class MyRobot(wpilib.TimedRobot):
+
                 def robotInit(self):
                     self.automodes = AutonomousModeSelector('autonomous')
-            
+
+                def autonomousInit(self):
+                    self.automodes.start()
+
                 def autonomousPeriodic(self):
-                    self.automodes.run()
-                    
-        
-        For ``SampleRobot``, for the simplest usage you would do this::
-        
-            class MyRobot(wpilib.SampleRobot):
-            
-                def robotInit(self):
-                    self.automodes = AutonomousModeSelector('autonomous')
-            
-                def autonomous(self):
-                    self.automodes.run()
-       
+                    self.automodes.periodic()
+
+                def disabledInit(self):
+                    self.automodes.disable()
+
         If you use AutonomousModeSelector, you may also be interested in
         the autonomous state machine helper (:class:`.StatefulAutonomous`).
         
@@ -65,6 +62,13 @@ class AutonomousModeSelector:
                   unit tests like so::
                   
                       from robotpy_ext.autonomous.selector_tests import *
+
+        .. note::
+
+           For your autonomous mode's ``on_disable`` method to be called,
+           you must call :meth:`disable` in ``disabledInit``.
+
+           It is okay to not call :meth:`disable` if you do not need ``on_disable``.
     """
 
     def __init__(self, autonomous_pkgname, *args, **kwargs):
@@ -118,11 +122,11 @@ class AutonomousModeSelector:
             #    on the field..
 
             for name, obj in inspect.getmembers(module, inspect.isclass):
-
-                if hasattr(obj, "MODE_NAME"):
+                mode_name = getattr(obj, "MODE_NAME", None)
+                if mode_name is not None:
 
                     # don't allow the driver to select this mode
-                    if hasattr(obj, "DISABLED") and obj.DISABLED:
+                    if getattr(obj, "DISABLED", False):
                         logger.warning(
                             "autonomous mode %s is marked as disabled", obj.MODE_NAME
                         )
@@ -137,16 +141,15 @@ class AutonomousModeSelector:
                         else:
                             continue
 
-                    if instance.MODE_NAME in self.modes:
+                    if mode_name in self.modes:
                         if not self.ds.isFMSAttached():
                             raise RuntimeError(
-                                "Duplicate name %s in %s"
-                                % (instance.MODE_NAME, module_filename)
+                                "Duplicate name %s in %s" % (mode_name, module_filename)
                             )
 
                         logger.error(
                             "Duplicate name %s specified by object type %s in module %s",
-                            instance.MODE_NAME,
+                            mode_name,
                             name,
                             module_filename,
                         )
@@ -166,23 +169,23 @@ class AutonomousModeSelector:
         logger.info("Loaded autonomous modes:")
         for k, v in sorted(self.modes.items()):
 
-            if hasattr(v, "DEFAULT") and v.DEFAULT == True:
+            if getattr(v, "DEFAULT", False):
                 logger.info(" -> %s [Default]", k)
-                self.chooser.addDefault(k, v)
+                self.chooser.setDefaultOption(k, v)
                 default_modes.append(k)
             else:
                 logger.info(" -> %s", k)
-                self.chooser.addObject(k, v)
+                self.chooser.addOption(k, v)
 
             mode_names.append(k)
 
         if len(self.modes) == 0:
             logger.warning("-- no autonomous modes were loaded!")
 
-        self.chooser.addObject("None", None)
+        self.chooser.addOption("None", None)
 
         if len(default_modes) == 0:
-            self.chooser.addDefault("None", None)
+            self.chooser.setDefaultOption("None", None)
         elif len(default_modes) != 1:
             if not self.ds.isFMSAttached():
                 raise RuntimeError(
@@ -198,13 +201,19 @@ class AutonomousModeSelector:
 
         logger.info("Autonomous switcher initialized")
 
-    def run(self, control_loop_wait_time=0.020, iter_fn=None, on_exception=None):
+    def run(
+        self,
+        control_loop_wait_time=0.020,
+        iter_fn=None,
+        on_exception=None,
+        watchdog: Union[wpilib.Watchdog, SimpleWatchdog] = None,
+    ):
         """
-            This function does everything required to implement autonomous
-            mode behavior. You should call this from your autonomous mode
-            function -- ``autonomousPeriodic`` in :class:`.IterativeRobot`,
-            or ``autonomous`` in :class:`.SampleRobot`.
-            
+            This method implements the entire autonomous loop.
+
+            Do not call this from ``TimedRobot`` as this will break the
+            timing of your control loop when your robot switches to teleop.
+
             This function will NOT exit until autonomous mode has ended. If
             you need to execute code in all autonomous modes, pass a function
             or list of functions as the ``iter_fn`` parameter, and they will be
@@ -215,12 +224,25 @@ class AutonomousModeSelector:
                             autonomous mode is executing
             :param on_exception: Called when an uncaught exception is raised,
                                  must take a single keyword arg "forceReport"
+            :param watchdog: a WPILib Watchdog to feed every iteration
         """
+        if watchdog:
+            watchdog.reset()
+
+            if isinstance(watchdog, SimpleWatchdog):
+                watchdog_check_expired = watchdog.printIfExpired
+            else:
+
+                def watchdog_check_expired():
+                    if watchdog.isExpired():
+                        watchdog.printEpochs()
 
         logger.info("Begin autonomous")
 
         if iter_fn is None:
-            iter_fn = lambda: None
+            iter_fn = (lambda: None,)
+        elif not isinstance(iter_fn, (list, tuple)):
+            iter_fn = (iter_fn,)
 
         if on_exception is None:
             on_exception = self._on_exception
@@ -233,37 +255,85 @@ class AutonomousModeSelector:
             self._on_autonomous_enable()
         except:
             on_exception(forceReport=True)
+        if watchdog:
+            watchdog.addEpoch("auto on_enable")
 
         #
         # Autonomous control loop
         #
 
+        observe = hal.observeUserProgramAutonomous
+
         with NotifierDelay(control_loop_wait_time) as delay:
-            while self.ds.isAutonomous() and self.ds.isEnabled():
+            while self.ds.isAutonomousEnabled():
+                observe()
                 try:
                     self._on_iteration(timer.get())
                 except:
                     on_exception()
+                if watchdog:
+                    watchdog.addEpoch("auto on_iteration")
 
-                if isinstance(iter_fn, (list, tuple)):
-                    for fn in iter_fn:
-                        fn()
-                else:
-                    iter_fn()
+                for fn in iter_fn:
+                    fn()
+
+                if watchdog:
+                    watchdog.addEpoch("robotPeriodic()")
+                    watchdog.disable()
+
+                    watchdog_check_expired()
 
                 delay.wait()
+                if watchdog:
+                    watchdog.reset()
 
         #
         # Done with autonomous, finish up
         #
 
         try:
-            self._on_autonomous_disable()
+            self.disable()
         except:
             on_exception(forceReport=True)
 
         logger.info("Autonomous mode ended")
 
+    def start(self) -> None:
+        """Start autonomous mode.
+
+        This initialises the selected autonomous mode.
+        Call this from your ``autonomousInit`` method.
+
+        .. versionadded:: 2020.1.5
+        """
+        self.timer = wpilib.Timer()
+        self.timer.start()
+
+        self._on_autonomous_enable()
+
+    def periodic(self) -> None:
+        """Execute one control loop iteration of the active autonomous mode.
+
+        Call this from your ``autonomousPeriodic`` method.
+
+        .. versionadded:: 2020.1.5
+        """
+        self._on_iteration(self.timer.get())
+
+    def disable(self) -> None:
+        """Disables the active autonomous mode.
+
+        You can call this from your ``disabledInit`` method
+        to call your autonomous mode's ``on_disable`` method.
+
+        .. versionadded:: 2020.1.5
+        """
+        if self.active_mode is not None:
+            logger.info("Disabling '%s'", self.active_mode.MODE_NAME)
+            self.active_mode.on_disable()
+
+        self.active_mode = None
+
     #
     #   Internal methods used to implement autonomous mode switching, and
     #   are called automatically
@@ -284,21 +354,13 @@ class AutonomousModeSelector:
             self.active_mode = self.chooser.getSelected()
 
         if self.active_mode is not None:
-            logger.info("Enabling '%s'" % self.active_mode.MODE_NAME)
+            logger.info("Enabling '%s'", self.active_mode.MODE_NAME)
             self.active_mode.on_enable()
         else:
             logger.warning(
                 "No autonomous modes were selected, not enabling autonomous mode"
             )
 
-    def _on_autonomous_disable(self):
-        """Disable the active autonomous mode"""
-        if self.active_mode is not None:
-            logger.info("Disabling '%s'" % self.active_mode.MODE_NAME)
-            self.active_mode.on_disable()
-
-        self.active_mode = None
-
     def _on_iteration(self, time_elapsed):
         """Run the code for the current autonomous mode"""
         if self.active_mode is not None:
